/*
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * File created by null on 4/12/09.
 */

#import "HSCController.h"
#import "VMRegion.h"
#import "SearchContext.h"
#import "Variable.h"
#import "HDLMaximumLengthTextFieldFormatter.h"
#import "HSCPreferencesController.h"
#import "NSAlertCheckbox.h"

#import <SecurityFoundation/SFAuthorization.h>
#import <Security/AuthorizationTags.h>
#import "unistd.h"

// networking stuff
#import <errno.h>
#import <sys/time.h>
#import <sys/types.h>
#import <sys/socket.h>
#import <netinet/in.h>
#import <arpa/inet.h>
#import <netdb.h>
#import <fcntl.h>

#pragma mark Authorization functions

// these authorization and permission functions are taken (and modified) from ihaxgamez and virtue desktops

void authMe(char * FullPathToMe)
{
	// get authorization as root
	
	OSStatus myStatus;
	
	// set up Authorization Item
	AuthorizationItem myItems[1];
	myItems[0].name = kAuthorizationRightExecute;
	myItems[0].valueLength = 0;
	myItems[0].value = NULL;
	myItems[0].flags = 0;
	
	// Set up Authorization Rights
	AuthorizationRights myRights;
	myRights.count = sizeof (myItems) / sizeof (myItems[0]);
	myRights.items = myItems;
	
	// set up Authorization Flags
	AuthorizationFlags myFlags;
	myFlags =
	kAuthorizationFlagDefaults |
	kAuthorizationFlagInteractionAllowed |
	kAuthorizationFlagExtendRights;
	
	// Create an Authorization Ref using Objects above. NOTE: Login bod comes up with this call.
	AuthorizationRef myAuthorizationRef;
	myStatus = AuthorizationCreate (&myRights, kAuthorizationEmptyEnvironment, myFlags, &myAuthorizationRef);
	
	if (myStatus == errAuthorizationSuccess)
	{
		// prepare communication path - used to signal that process is loaded
		FILE *myCommunicationsPipe = NULL;
		char myReadBuffer[] = " ";
		
		// run this app in GOD mode by passing authorization ref and comm pipe (asynchoronous call to external application)
		myStatus = AuthorizationExecuteWithPrivileges(myAuthorizationRef,FullPathToMe,kAuthorizationFlagDefaults,nil,&myCommunicationsPipe);
		
		// external app is running asynchronously - it will send to stdout when loaded
		if (myStatus == errAuthorizationSuccess)
		{
			read (fileno (myCommunicationsPipe), myReadBuffer, sizeof (myReadBuffer));
			fclose(myCommunicationsPipe);
		}
		
		// release authorization reference
		myStatus = AuthorizationFree (myAuthorizationRef, kAuthorizationFlagDestroyRights);
	}
}

bool checkExecutablePermissions(void)
{
	NSDictionary	*applicationAttributes = [[NSFileManager defaultManager] fileAttributesAtPath:[[NSBundle mainBundle] executablePath] traverseLink: YES];
	
	// We expect 2755 as octal (1517 as decimal, -rwxr-sr-x as extended notation)
	return ([applicationAttributes filePosixPermissions] == 1517 && [[applicationAttributes fileGroupOwnerAccountName] isEqualToString: @"procmod"]);
}

bool amIWorthy(void)
{
	// running as root?
	AuthorizationRef myAuthRef;
	OSStatus stat = AuthorizationCopyPrivilegedReference(&myAuthRef,kAuthorizationFlagDefaults);
	
	return stat == errAuthorizationSuccess || checkExecutablePermissions();
}

@interface HSCController (PrivateAPI)
- (void)readServerInfo:(NSTimer *)timer;
- (void)sendServerInfoToAllClients;
- (void)sendServerOffMessageToAllClients;
- (void)openPreferencesAndTellServerFailedToLaunch;
@end

static HSCController *sharedController = nil;

@implementation HSCController

NSString *HSCServerNameAddressKey =						@"HSCServerNameAddressKey";
NSString *HSCServerShutOffNotification =				@"HSCServerShutOffNotification";
NSString *HSCServerFailedToStartNotification =			@"HSCServerFailedToStartNotification";
NSString *HSCShouldShowRestartingHaloServerWarning =	@"HSCShouldShowRestartingHaloServerWarning";
NSString *HSCShouldCacheRestartingHaloServerAddresses = @"HSCShouldCacheRestartingHaloServerAddresses";
NSString *HSCShouldMakeHostSpectatorWarning =			@"HSCShouldMakeHostSpectatorWarning";

#define PASSWORD_ADDRESS_OFFSET_FROM_SERVER_ADDRESS		0x9F4
#define MAX_PLAYERS_ADDRESS_OFFSET_FROM_SERVER_ADDRESS	0x19C
#define MAX_SERVER_NAME_LENGTH							31
#define MAX_PASSWORD_LENGTH								8
// 27000 == 15 minutes in halo time (30 = 1 second)
#define DEFAULT_TIME_LIMIT								27000
#define VERY_SMALL_TIME_LIMIT							1
#define HOST_TEAM_NUMBER_ADDRESS						0x4BD7AFEE
#define RED_TEAM_NUMBER									0
#define BLUE_TEAM_NUMBER								1
#define NON_EXISTENT_TEAM_NUMBER						2
#define SPECTATOR_WRITE_TIME_INTERVAL					1

#define APP_SUPPORT_PATH								[[NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask, YES) objectAtIndex:0] stringByAppendingString:@"/Application Support/HaloServerChanger"]
#define HALO_RESTART_ADDRESSES_PATH						[APP_SUPPORT_PATH stringByAppendingString:@"/halo_restart_addresses.hschistory"]

#pragma mark Useful functions

- (void)writeUnicodeStringFromString:(NSString *)string
							 address:(vm_address_t)address
					  requiredLength:(int)requiredLength
{
	int unicodeLength = [string length];
	unichar unicodeCharacters[unicodeLength];
	[string getCharacters:unicodeCharacters];
	
	int unicodeCharacterIndex;
	for (unicodeCharacterIndex = 0; unicodeCharacterIndex < unicodeLength; unicodeCharacterIndex++)
	{
		unicodeCharacters[unicodeCharacterIndex] = CFSwapInt16BigToHost(unicodeCharacters[unicodeCharacterIndex]);
	}
	
	VMWriteBytes(haloProcessID, address, unicodeCharacters, unicodeLength * sizeof(unichar));
	
	if (requiredLength)
	{
		// max length - length of string + zero terminator
		int numberOfBytesToWrite = requiredLength - unicodeLength + 1;
		
		int bytesIndex;
		for (bytesIndex = 0; bytesIndex < numberOfBytesToWrite; bytesIndex++)
		{
			unichar zero = 0;
			VMWriteBytes(haloProcessID, address + unicodeLength * sizeof(unichar) + bytesIndex, &zero, sizeof(unichar));
		}
	}
	else
	{
		unichar zero = 0;
		VMWriteBytes(haloProcessID, address + unicodeLength * sizeof(unichar), &zero, sizeof(unichar));
	}
}

- (NSString *)readUnicodeStringWithLength:(vm_size_t)length
								  address:(vm_address_t)address
{
	unichar unicodeCharacters[length];
	vm_size_t size = length * sizeof(unichar);
	VMReadBytes(haloProcessID, address, &unicodeCharacters, &size);
	
	int unicodeCharacterIndex;
	for (unicodeCharacterIndex = 0; unicodeCharacterIndex < length && unicodeCharacters[unicodeCharacterIndex] != 0; unicodeCharacterIndex++)
	{
		unicodeCharacters[unicodeCharacterIndex] = CFSwapInt16BigToHost(unicodeCharacters[unicodeCharacterIndex]);
	}
	
	return [NSString stringWithCharacters:unicodeCharacters
								   length:unicodeCharacterIndex];
}

- (short)readInMaxPlayers
{
	short maxPlayers;
	
	vm_size_t size = sizeof(short);
	VMReadBytes(haloProcessID, _maxPlayersAddress, &maxPlayers, &size);
	
	return CFSwapInt16BigToHost(maxPlayers);
}

- (NSString *)readInServerName
{
	NSString *serverName = [self readUnicodeStringWithLength:MAX_SERVER_NAME_LENGTH
													 address:_serverNameAddress];
	
	return [serverName isEqualToString:@""] ? @"HALO SERVER" : serverName;
}

- (NSString *)readInPassword
{
	return [self readUnicodeStringWithLength:MAX_PASSWORD_LENGTH
									 address:_passwordAddress];
}

+ (HSCController *)sharedController
{
	return sharedController;
}

#pragma mark Birth

+ (void)initialize
{
	[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObject:[NSNumber numberWithInt:0]
																						forKey:HSCServerNameAddressKey]];
	
	[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObject:[NSNumber numberWithInt:NSOffState]
																						forKey:HSCEnableRemoteAccess]];
	
	[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObject:@""
																						forKey:HSCServerPassword]];
	
	const int DEFAULT_PORT = 2395;
	[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObject:[NSNumber numberWithInt:DEFAULT_PORT]
																						forKey:HSCListeningPort]];
	
	[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObject:[NSNumber numberWithBool:YES]
																						forKey:HSCShouldShowRestartingHaloServerWarning]];
	
	[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObject:[NSNumber numberWithBool:YES]
																						forKey:HSCShouldCacheRestartingHaloServerAddresses]];
	
	[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObject:[NSNumber numberWithBool:YES]
																						forKey:HSCShouldMakeHostSpectatorWarning]];
}

#pragma mark Finding Process ID

- (void)assignHaloProcessIDFromApplicationDictionary:(NSDictionary *)applicationDictionary
{
	if ([[applicationDictionary objectForKey:@"NSApplicationName"] rangeOfString:@"Halo Demo"
																		 options:NSCaseInsensitiveSearch].location != NSNotFound)
	{
		haloProcessID = [[applicationDictionary objectForKey:@"NSApplicationProcessIdentifier"] intValue];
		[syncButton setEnabled:YES];
		
		if (!readServerInfoTimer && ![setupWindow isVisible])
		{
			readServerInfoTimer = [NSTimer scheduledTimerWithTimeInterval:1.0
																   target:self
																 selector:@selector(readServerInfo:)
																 userInfo:nil
																  repeats:YES];
		}
	}
}

- (void)applicationDidLaunch:(NSNotification *)notification
{
	[self assignHaloProcessIDFromApplicationDictionary:[notification userInfo]];
}

- (void)applicationDidTerminate:(NSNotification *)notification
{
	if ([[[notification userInfo] objectForKey:@"NSApplicationName"] rangeOfString:@"Halo Demo"
																		   options:NSCaseInsensitiveSearch].location != NSNotFound)
	{
		haloProcessID = 0;
		
		if (![[[NSUserDefaults standardUserDefaults] objectForKey:HSCServerNameAddressKey] intValue])
		{
			[syncButton setEnabled:NO];
		}
		else
		{
			[readServerInfoTimer invalidate];
			readServerInfoTimer = nil;
			[self readServerInfo:nil];
		}
	}
}

#pragma mark Delegates

- (void)applicationDidFinishLaunching:(NSNotification *)notification
{
	BOOL needsAuthorizationCheck = YES;
#ifdef __ppc__
	// PPC machines whose operating system is below leopard do not need authorization
	SInt32 osxMinorVersion;
	Gestalt(gestaltSystemVersionMinor, &osxMinorVersion);
	if (osxMinorVersion < 5)
	{
		needsAuthorizationCheck = NO;
	}
#endif
	
	if (needsAuthorizationCheck)
	{
		if (amIWorthy())
		{
			// when the app restarts from asking the user his adminstator's password, the app isn't activated for some reason, so we'll activate it
			[NSApp activateIgnoringOtherApps:YES];
			
#ifndef _DEBUG
			// if we don't print this, the old instance of the app will stll be alive. this behavior is wanted in debug mode because xcode's debugger will remain opened
			printf("Don't forget to flush! ;-) "); // signal back to close caller
#endif
			fflush(stdout);
		}
		else
		{
			extern char *gExecutablePath;
			authMe(gExecutablePath);
			[NSApp terminate:nil];
		}
	}
	
	sharedController = self;
	
	if ([[NSUserDefaults standardUserDefaults] integerForKey:HSCEnableRemoteAccess])
	{
		[self enableRemoteAccess:YES];
	}
	
	if ([[[NSUserDefaults standardUserDefaults] objectForKey:HSCServerNameAddressKey] intValue])
	{
		_serverNameAddress = [[[NSUserDefaults standardUserDefaults] objectForKey:HSCServerNameAddressKey] intValue];
		_passwordAddress = _serverNameAddress + PASSWORD_ADDRESS_OFFSET_FROM_SERVER_ADDRESS;
		_maxPlayersAddress = _serverNameAddress + MAX_PLAYERS_ADDRESS_OFFSET_FROM_SERVER_ADDRESS;
		
		[mainWindow makeKeyAndOrderFront:nil];
		
		readServerInfoTimer = [NSTimer scheduledTimerWithTimeInterval:1.0
															   target:self
															 selector:@selector(readServerInfo:)
															 userInfo:nil
															  repeats:YES];
	}
	else
	{
		[setupWindow makeKeyAndOrderFront:nil];
	}
	
	[mainWindow setExcludedFromWindowsMenu:YES];
	
	[serverNameTextField setFormatter:[HDLMaximumLengthTextFieldFormatter formatterWithMaximumLength:MAX_SERVER_NAME_LENGTH]];
	[passwordTextField setFormatter:[HDLMaximumLengthTextFieldFormatter formatterWithMaximumLength:MAX_PASSWORD_LENGTH]];
	
	NSEnumerator *launchedApplicationsEnumerator = [[[NSWorkspace sharedWorkspace] launchedApplications] objectEnumerator];
	NSDictionary *applicationDictionary;
	
	while ((applicationDictionary = [launchedApplicationsEnumerator nextObject]))
	{
		[self assignHaloProcessIDFromApplicationDictionary:applicationDictionary];
	}
	
	[[[NSWorkspace sharedWorkspace] notificationCenter] addObserver:self
														   selector:@selector(applicationDidLaunch:)
															   name:NSWorkspaceDidLaunchApplicationNotification
															 object:nil];
	[[[NSWorkspace sharedWorkspace] notificationCenter] addObserver:self
														   selector:@selector(applicationDidTerminate:)
															   name:NSWorkspaceDidTerminateApplicationNotification
															 object:nil];
	
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(maxPlayersTextDidChange:)
												 name:NSTextDidChangeNotification
											   object:nil];
}

#pragma mark Menu Item Validation

- (BOOL)validateMenuItem:(NSMenuItem *)theMenuItem
{
	if ([theMenuItem action] == @selector(showResynchronizationWindow:))
	{
		return ![setupWindow isVisible];
	}
	
	else if ([theMenuItem action] == @selector(restartHaloDemoServer:))
	{
		return serverIsUp && ![setupWindow isVisible];
	}
	
	else if ([theMenuItem action] == @selector(makeHostASpectator:) && !hostIsGoingToSpectateFromThreadNotice)
	{
		return serverIsUp && ![setupWindow isVisible];
	}
	
	else if ([theMenuItem action] == @selector(openMainWindow:))
	{
		[theMenuItem setState:[mainWindow isVisible]];
		return ![setupWindow isVisible];
	}
	else if ([theMenuItem action] == @selector(removeHaloGameRestartCache:))
	{
		return [[NSFileManager defaultManager] fileExistsAtPath:HALO_RESTART_ADDRESSES_PATH];
	}
	
	return YES;
}

#pragma mark Actions

- (IBAction)applyChanges:(id)sender
{
	// server name
	
	[self writeUnicodeStringFromString:[serverNameTextField stringValue]
							   address:_serverNameAddress
						requiredLength:0];
	
	if ([[serverNameTextField stringValue] isEqualToString:@""])
	{
		[serverNameTextField setStringValue:@"HALO SERVER"];
	}
	
	// password
	
	[self writeUnicodeStringFromString:[passwordTextField stringValue]
							   address:_passwordAddress
						requiredLength:MAX_PASSWORD_LENGTH];
	
	// max players
	
	short maxPlayers = (short)[maxPlayersStepper intValue];
	[maxPlayersTextField setIntValue:maxPlayers];
	maxPlayers = CFSwapInt16HostToBig(maxPlayers);
	VMWriteBytes(haloProcessID, _maxPlayersAddress, &maxPlayers, sizeof(short));
	
	if (remoteAccessServerIsUp)
	{
		[self sendServerInfoToAllClients];
	}
}

- (IBAction)restoreServerInfo:(id)sender
{
	short maxPlayers = [self readInMaxPlayers];
	
	[maxPlayersStepper setIntValue:maxPlayers];
	[maxPlayersTextField setIntValue:maxPlayers];
	
	[serverNameTextField setStringValue:[self readInServerName]];
	
	[passwordTextField setStringValue:[self readInPassword]];
}

- (void)readServerInfo:(NSTimer *)timer
{
	if ([setupWindow isVisible])
	{
		return;
	}
	
	if (hostIsGoingToSpectateFromThreadNotice)
	{
		[self makeHostASpectator:makeHostASpectatorMenuItem];
		hostIsGoingToSpectateFromThreadNotice = NO;
	}
	
	if (!haloProcessID)
	{
		if (remoteAccessServerIsUp && serverIsUp)
		{
			[self sendServerOffMessageToAllClients];
		}
		
		serverIsUp = NO;
		[applyButton setEnabled:NO];
		[restoreButton setEnabled:NO];
		
		[makeHostASpectatorTimer invalidate];
		makeHostASpectatorTimer = nil;
		
		[makeHostASpectatorMenuItem setState:NSOffState];
	}
	else
	{
		// read in max players and if it is not 0, then the server will be up
		short maxPlayers = [self readInMaxPlayers];
		
		if (maxPlayers && !serverIsUp)
		{
			[maxPlayersTextField setIntValue:maxPlayers];
			[maxPlayersStepper setIntValue:maxPlayers];
			
			[serverNameTextField setStringValue:[self readInServerName]];
			
			[passwordTextField setStringValue:[self readUnicodeStringWithLength:MAX_PASSWORD_LENGTH
																		address:_passwordAddress]];
			
			serverIsUp = YES;
			[applyButton setEnabled:YES];
			[restoreButton setEnabled:YES];
			
			short hostTeamNumber;
			vm_size_t hostTeamNumberSize = sizeof(short);
			VMReadBytes(haloProcessID, HOST_TEAM_NUMBER_ADDRESS, &hostTeamNumber, &hostTeamNumberSize);
			hostTeamNumber = CFSwapInt16BigToHost(hostTeamNumber);
			
			if (hostTeamNumber != RED_TEAM_NUMBER && hostTeamNumber != BLUE_TEAM_NUMBER)
			{
				makeHostASpectatorTimer = [NSTimer scheduledTimerWithTimeInterval:SPECTATOR_WRITE_TIME_INTERVAL
																		   target:self
																		 selector:@selector(makeHostSpectate:)
																		 userInfo:nil
																		  repeats:YES];
				[makeHostASpectatorMenuItem setState:NSOnState];
			}
			
			if (remoteAccessServerIsUp)
			{
				[self sendServerInfoToAllClients];
			}
		}
		else if (!maxPlayers && serverIsUp)
		{
			if (remoteAccessServerIsUp)
			{
				[self sendServerOffMessageToAllClients];
			}
			
			serverIsUp = NO;
			[applyButton setEnabled:NO];
			[restoreButton setEnabled:NO];
			
			[makeHostASpectatorTimer invalidate];
			makeHostASpectatorTimer = nil;
			
			[makeHostASpectatorMenuItem setState:NSOffState];
		}
	}
}

- (IBAction)maxPlayersStepper:(id)sender
{
	[maxPlayersTextField setIntValue:[sender intValue]];
	[maxPlayersTextField selectText:nil];
}

- (void)maxPlayersTextDidChange:(NSNotification *)aNotification
{
	// we want the text field to be continuous
	
	if ([aNotification object] == [maxPlayersTextField currentEditor])
	{
		if ([maxPlayersTextField intValue] > 1 && [maxPlayersTextField intValue] <= 16)
		{
			[maxPlayersStepper setIntValue:[maxPlayersTextField intValue]];
		}
	}
}

// if sender is nil, it won't bother trying to display the alert notice
- (IBAction)restartHaloDemoServer:(id)sender
{
	BOOL userWantsToRestart = YES;
	if (sender && [[NSUserDefaults standardUserDefaults] boolForKey:HSCShouldShowRestartingHaloServerWarning])
	{
		NSAlertCheckbox *alert = [NSAlertCheckbox alertWithMessageText:@"Do you want to continue?"
														 defaultButton:@"Yes"
													   alternateButton:@"No"
														   otherButton:nil
													   informativeText:@"This may only work with the standard gametypes Slayer and CTF. Other gametypes may not work unless their time limit is set at 15 minutes. This could cause harm on other gametypes with a different time limit."];
		[alert setShowsCheckbox:YES];
		[alert setCheckboxText:@"Don't ask me again."];
		
		if ([alert runModal] == NSAlertAlternateReturn)
		{
			userWantsToRestart = NO;
		}
		else if ([alert checkboxState] == NSOnState)
		{
			[[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithBool:NO]
													  forKey:HSCShouldShowRestartingHaloServerWarning];
		}
	}
	
	if (userWantsToRestart)
	{
		if (![[NSFileManager defaultManager] fileExistsAtPath:APP_SUPPORT_PATH])
		{
			[[NSFileManager defaultManager] createDirectoryAtPath:APP_SUPPORT_PATH
													   attributes:nil];
		}
		
		Variable *value = [[Variable alloc] init];
		
		if ([[NSUserDefaults standardUserDefaults] boolForKey:HSCShouldCacheRestartingHaloServerAddresses] && [[NSFileManager defaultManager] fileExistsAtPath:HALO_RESTART_ADDRESSES_PATH])
		{
			NSArray *haloRestartServerAddressesArray = [NSKeyedUnarchiver unarchiveObjectWithFile:HALO_RESTART_ADDRESSES_PATH];
			NSEnumerator *haloRestartServerAddressesEnumerator = [haloRestartServerAddressesArray objectEnumerator];
			NSNumber *haloRestartServerAddress;
			
			int timeLimit = CFSwapInt32HostToBig(VERY_SMALL_TIME_LIMIT);
			while (haloRestartServerAddress = [haloRestartServerAddressesEnumerator nextObject])
			{
				VMWriteBytes(haloProcessID, [haloRestartServerAddress intValue], &timeLimit, sizeof(int));
			}
			
			if (remoteAccessServerIsUp && interestedClientSocket)
			{
				send(interestedClientSocket, "he\n", 3, 0);
				interestedClientSocket = 0;
			}
			
			[NSTimer scheduledTimerWithTimeInterval:2
											 target:self
										   selector:@selector(restoreTimeLimitAddressesFromAddressFile:)
										   userInfo:[haloRestartServerAddressesArray retain]
											repeats:NO];
		}
		else
		{
			int timeLimit = CFSwapInt32HostToBig(DEFAULT_TIME_LIMIT);
			[value setValue:&timeLimit
					   size:sizeof(int)];
			
			SearchContext *context = [[SearchContext alloc] initWithPID:haloProcessID
														 searchOperator:TCEqual
																  value:value];
			
			restartingHaloDemoTask = [[ThreadedTask alloc] initWithFunction:[context iterationFunction]
																	context:context
																   delegate:self];
			
			[context release];
			
			[restartingHaloServerProgressIndicator setDoubleValue:0];
			restartingHaloServerSession = [NSApp beginModalSessionForWindow:restartingHaloServerWindow];
			
			[restartingHaloDemoTask run];
		}
		
		[value release];
	}
}

- (IBAction)cancelRestartingHaloDemoServer:(id)sender
{
	if ([restartingHaloDemoTask isRunning])
	{
		if (remoteAccessServerIsUp && interestedClientSocket)
		{
			send(interestedClientSocket, "hc\n", 3, 0);
			interestedClientSocket = 0;
		}
		
		[NSApp endModalSession:restartingHaloServerSession];
		
		[restartingHaloDemoTask cancel];
		[restartingHaloDemoTask release];
		restartingHaloDemoTask = nil;
		
		[restartingHaloServerWindow orderOut:nil];
		restartingHaloServerSession = NULL;
	}
}

- (void)restoreTimeLimitAddressesFromAddressFile:(NSTimer *)timer
{
	NSEnumerator *haloRestartServerAddressesEnumerator = [[timer userInfo] objectEnumerator];
	NSNumber *haloRestartServerAddress;
	
	int timeLimit = CFSwapInt32HostToBig(DEFAULT_TIME_LIMIT);
	while (haloRestartServerAddress = [haloRestartServerAddressesEnumerator nextObject])
	{
		VMWriteBytes(haloProcessID, [haloRestartServerAddress intValue], &timeLimit, sizeof(int));
	}
	
	[[timer userInfo] release];
}

- (void)restoreTimeLimitAddresses:(NSTimer *)timer
{
	ThreadedTask *restartHaloTask = [timer userInfo];
	
	TCArray addresses = ((SearchContext *)[restartHaloTask context])->addresses;
	
	unsigned addressIndex;
	for (addressIndex = 0; addressIndex < TCArrayElementCount(addresses); addressIndex++)
	{
		int address = *((int *)TCArrayElementAtIndex(addresses, addressIndex));
		int newTimeLimit = CFSwapInt32HostToBig(DEFAULT_TIME_LIMIT);
		VMWriteBytes(haloProcessID, address, &newTimeLimit, sizeof(int));
	}
	
	[restartingHaloDemoTask release];
	restartingHaloDemoTask = nil;
}

- (IBAction)makeHostASpectator:(id)sender
{
	if ([sender state] == NSOffState)
	{
		BOOL userWantsHostToSpectate = YES;
		
		if (!remoteAccessServerIsUp && [[NSUserDefaults standardUserDefaults] boolForKey:HSCShouldMakeHostSpectatorWarning])
		{
			NSAlertCheckbox *alert = [NSAlertCheckbox alertWithMessageText:@"Do you want to continue?"
															 defaultButton:@"Yes"
														   alternateButton:@"No"
															   otherButton:nil
														   informativeText:@"This will not work in Slayer and will only work in team based gametypes."];
			[alert setShowsCheckbox:YES];
			[alert setCheckboxText:@"Don't ask me again."];
			
			if ([alert runModal] == NSAlertAlternateReturn)
			{
				userWantsHostToSpectate = NO;
			}
			else if ([alert checkboxState] == NSOnState)
			{
				[[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithBool:NO]
														  forKey:HSCShouldMakeHostSpectatorWarning];
			}
		}
		
		if (userWantsHostToSpectate)
		{
			[sender setState:NSOnState];
			
			makeHostASpectatorTimer = [NSTimer scheduledTimerWithTimeInterval:SPECTATOR_WRITE_TIME_INTERVAL
																	   target:self
																	 selector:@selector(makeHostSpectate:)
																	 userInfo:nil
																	  repeats:YES];
		}
	}
	else if ([sender state] == NSOnState)
	{
		[makeHostASpectatorTimer invalidate];
		makeHostASpectatorTimer = nil;
		
		[sender setState:NSOffState];
		
		short teamNumber = CFSwapInt16HostToBig(RED_TEAM_NUMBER);
		VMWriteBytes(haloProcessID, HOST_TEAM_NUMBER_ADDRESS, &teamNumber, sizeof(short));
	}
	
	if (remoteAccessServerIsUp)
	{
		[self sendServerInfoToAllClients];
	}
}

- (void)makeHostSpectate:(NSTimer *)timer
{
	if (haloProcessID && serverIsUp)
	{
		unsigned short currentTeamNumber;
		vm_size_t size = sizeof(short);
		VMReadBytes(haloProcessID, HOST_TEAM_NUMBER_ADDRESS, &currentTeamNumber, &size);
		currentTeamNumber = CFSwapInt16BigToHost(currentTeamNumber);
		
		if (currentTeamNumber != NON_EXISTENT_TEAM_NUMBER)
		{
			// if the host is alive, move the host's player position so he automatically dies
			
			const int hostObjectIDAddress = 0x4BD7B002;
			const int firstTableObjectArrayAddress = 0x4BB206EC;
			const int tableObjectArraySize = 12;
			const unsigned short invalidHostObjectID = 0xFFFF;
			const int offsetToObjectArrayTablePointer = 0x8;
			
			unsigned short hostObjectID;
			size = sizeof(short);
			VMReadBytes(haloProcessID, hostObjectIDAddress, &hostObjectID, &size);
			hostObjectID = CFSwapInt16BigToHost(hostObjectID);
			
			if (hostObjectID != 0 && hostObjectID != invalidHostObjectID)
			{
				// the host is alive
				
				unsigned int haloObjectPointerAddress = firstTableObjectArrayAddress + hostObjectID * tableObjectArraySize + offsetToObjectArrayTablePointer;
				
				unsigned int haloObjectPointer;
				size = sizeof(int);
				VMReadBytes(haloProcessID, haloObjectPointerAddress, &haloObjectPointer, &size);
				haloObjectPointer = CFSwapInt32BigToHost(haloObjectPointer);
				
				const int offsetToPlayerXCoordinate = 0x5C;
				
				// Kill the host!
				float newHostXValue = 1000.0;
				int *newHostXValuePointer = (int *)&newHostXValue;
				*newHostXValuePointer = CFSwapInt32HostToBig(*((int *)&newHostXValue));
				VMWriteBytes(haloProcessID, haloObjectPointer + offsetToPlayerXCoordinate, &newHostXValue, sizeof(float));
			}
			
			short teamNumber = CFSwapInt16HostToBig(NON_EXISTENT_TEAM_NUMBER);
			VMWriteBytes(haloProcessID, HOST_TEAM_NUMBER_ADDRESS, &teamNumber, sizeof(short));
		}
	}
}

- (IBAction)showResynchronizationWindow:(id)sender
{	
	[readServerInfoTimer invalidate];
	readServerInfoTimer = nil;
	
	[mainWindow orderOut:nil];
	
	[syncButton setTitle:@"Sync"];
	[syncButton setKeyEquivalent:@"\r"];
	[syncButton setEnabled:YES];
	
	[setupWindow makeKeyAndOrderFront:nil];
}

- (BOOL)fileManager:(NSFileManager *)manager shouldProceedAfterError:(NSDictionary *)errorInfo
{
	NSLog(@"Path: %@", [errorInfo objectForKey:@"Path"]);
	NSLog(@"toPath: %@", [errorInfo objectForKey:@"ToPath"]);
	NSLog(@"Error: %@", [errorInfo objectForKey:@"Error"]);
	
	return NO;
}

- (IBAction)removeHaloGameRestartCache:(id)sender
{
	[[NSFileManager defaultManager] removeFileAtPath:HALO_RESTART_ADDRESSES_PATH
											 handler:self];
}

#pragma mark Syncing

- (void)threadedTaskFinished:(ThreadedTask *)theTask
{
	if (theTask == syncingTask)
	{
		int validServerNameAddress = 0;
		
		TCArray addresses = ((SearchContext *)[theTask context])->addresses;
		unsigned addressIndex;
		for (addressIndex = 0; addressIndex < TCArrayElementCount(addresses); addressIndex++)
		{
			int serverNameAddress = *((int *)TCArrayElementAtIndex(addresses, addressIndex));
			int passwordAddress = serverNameAddress + PASSWORD_ADDRESS_OFFSET_FROM_SERVER_ADDRESS;
			int maxPlayersAddress = serverNameAddress + MAX_PLAYERS_ADDRESS_OFFSET_FROM_SERVER_ADDRESS;
			
			NSString *password = [self readUnicodeStringWithLength:4
														   address:passwordAddress];
			
			short maxPlayers;
			vm_size_t size = sizeof(short);
			VMReadBytes(haloProcessID, maxPlayersAddress, &maxPlayers, &size);
			maxPlayers = CFSwapInt16BigToHost(maxPlayers);
			
			if ([password isEqualToString:@"pass"] && maxPlayers == 8)
			{
				validServerNameAddress = serverNameAddress;
				
				break;
			}
		}
		
		[syncingTask release];
		syncingTask = nil;
		
		[syncProgressIndicator setHidden:YES];
		[syncProgressIndicator setDoubleValue:0];
		
		if (!validServerNameAddress)
		{
			NSBeep();
			
			if (haloProcessID)
			{
				NSRunAlertPanel(@"Synchronization Failure",
								@"Are you sure you have a server running with a name of Q), password of pass, and max players set to 8?",
								@"OK", nil, nil);
				
				[syncButton setEnabled:YES];
				[syncButton setTitle:@"Sync"];
				[syncButton setKeyEquivalent:@"\r"];
			}
			else
			{
				[syncButton setTitle:@"Sync"];
				[syncButton setKeyEquivalent:@"\r"];
				
				NSRunAlertPanel(@"Synchronization Failure",
								@"Halo Demo has quit in the middle of the synchronization process.",
								@"OK", nil, nil);
			}
		}
		else
		{
			_serverNameAddress = validServerNameAddress;
			_passwordAddress = _serverNameAddress + PASSWORD_ADDRESS_OFFSET_FROM_SERVER_ADDRESS;
			_maxPlayersAddress = _serverNameAddress + MAX_PLAYERS_ADDRESS_OFFSET_FROM_SERVER_ADDRESS;
			
			[[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithInt:validServerNameAddress]
													  forKey:HSCServerNameAddressKey];
			
			[setupWindow orderOut:nil];
			[self readServerInfo:nil];
			[self restoreServerInfo:nil];
			[mainWindow makeKeyAndOrderFront:nil];
			
			if (!readServerInfoTimer)
			{
				readServerInfoTimer = [NSTimer scheduledTimerWithTimeInterval:1.0
																	   target:self
																	 selector:@selector(readServerInfo:)
																	 userInfo:nil
																	  repeats:YES];
			}
		}
	}
	else if (theTask == restartingHaloDemoTask)
	{
		TCArray addresses = ((SearchContext *)[theTask context])->addresses;
		// sanity check
		if (TCArrayElementCount(addresses) <= 30)
		{
			NSMutableArray *haloRestartAddressesArray = nil;
			if (![[NSFileManager defaultManager] fileExistsAtPath:HALO_RESTART_ADDRESSES_PATH])
			{
				haloRestartAddressesArray = [[NSMutableArray alloc] init];
			}
			unsigned addressIndex;
			for (addressIndex = 0; addressIndex < TCArrayElementCount(addresses); addressIndex++)
			{
				int address = *((int *)TCArrayElementAtIndex(addresses, addressIndex));
				int newTimeLimit = CFSwapInt32HostToBig(VERY_SMALL_TIME_LIMIT);
				VMWriteBytes(haloProcessID, address, &newTimeLimit, sizeof(int));
				
				[haloRestartAddressesArray addObject:[NSNumber numberWithInt:address]];
			}
			
			if (haloRestartAddressesArray)
			{
				if ([[NSUserDefaults standardUserDefaults] boolForKey:HSCShouldCacheRestartingHaloServerAddresses] && ![NSKeyedArchiver archiveRootObject:haloRestartAddressesArray
																																				   toFile:HALO_RESTART_ADDRESSES_PATH])
				{
					NSLog(@"Failure to save halo_restart_addresses.hschistory");
				}
				
				[haloRestartAddressesArray release];
			}
			
			if (remoteAccessServerIsUp && interestedClientSocket)
			{
				send(interestedClientSocket, "he\n", 3, 0);
				interestedClientSocket = 0;
			}
			
			[NSTimer scheduledTimerWithTimeInterval:2
											 target:self
										   selector:@selector(restoreTimeLimitAddresses:)
										   userInfo:theTask
											repeats:NO];
		}
		else
		{
			if (!remoteAccessServerIsUp)
			{
				NSBeep();
				NSRunAlertPanel(@"Restart Failure",
								@"The Halo server has failed to restart.",
								@"OK", nil, nil);
			}
			else if (remoteAccessServerIsUp && interestedClientSocket)
			{
				send(interestedClientSocket, "hf\n", 3, 0);
				interestedClientSocket = 0;
			}
			
			[restartingHaloDemoTask release];
			restartingHaloDemoTask = nil;
		}
		
		[NSApp endModalSession:restartingHaloServerSession];
		[restartingHaloServerWindow orderOut:nil];
		restartingHaloServerSession = NULL;
		
		restartingHaloDemoTaskShouldCancel = NO;
	}
}

- (void)threadedTaskCancelled:(ThreadedTask *)theTask
{
	if (theTask == syncingTask)
	{
		[syncProgressIndicator setHidden:YES];
		[syncProgressIndicator setDoubleValue:0];
		
		[syncButton setTitle:@"Sync"];
		[syncButton setKeyEquivalent:@"\r"];
		[syncButton setEnabled:YES];
		
		[syncingTask release];
		syncingTask = nil;
	}
	else if (theTask == restartingHaloDemoTask)
	{
		[NSApp endModalSession:restartingHaloServerSession];
		[restartingHaloServerWindow orderOut:nil];
		restartingHaloServerSession = NULL;
		
		[restartingHaloDemoTask release];
		restartingHaloDemoTask = nil;
		
		restartingHaloDemoTaskShouldCancel = NO;
	}
}

- (void)threadedTask:(ThreadedTask *)theTask failedWithErrorCode:(int)errorCode
{
	if (theTask == syncingTask)
	{
		[syncProgressIndicator setHidden:YES];
		[syncProgressIndicator setDoubleValue:0];
		
		NSBeep();
		NSRunAlertPanel(@"Synchronization Failure",
						@"Synchronization to Halo Demo has failed with an error code of %i.",
						@"OK", nil, nil, errorCode);
		
		NSLog(@"Error code: %i", errorCode);
		
		[syncButton setTitle:@"Sync"];
		[syncButton setKeyEquivalent:@"\r"];
		[syncButton setEnabled:YES];
		
		[syncingTask release];
		syncingTask = nil;
	}
	else if (theTask == restartingHaloDemoTask)
	{
		[NSApp endModalSession:restartingHaloServerSession];
		[restartingHaloServerWindow orderOut:nil];
		restartingHaloServerSession = NULL;
		
		NSLog(@"Error code is %i", errorCode);
		
		if (!remoteAccessServerIsUp)
		{
			NSBeep();
			NSRunAlertPanel(@"Restart Failure",
							@"Restarting Halo Demo has failed with an error code of %i.",
							@"OK", nil, nil, errorCode);
		}
		else if (remoteAccessServerIsUp && interestedClientSocket)
		{
			NSString *message = [NSString stringWithFormat:@"ha%i\n", errorCode];
			send(interestedClientSocket, [message UTF8String], strlen([message UTF8String]), 0);
			interestedClientSocket = 0;
		}
		
		[restartingHaloDemoTask release];
		restartingHaloDemoTask = nil;
		
		restartingHaloDemoTaskShouldCancel = NO;
	}
}

- (void)threadedTask:(ThreadedTask *)theTask reportedProgress:(int)theProgress
{
	if (theTask == syncingTask)
	{
		[syncProgressIndicator setDoubleValue:theProgress];
	}
	else if (theTask == restartingHaloDemoTask)
	{
		if (restartingHaloDemoTaskShouldCancel && remoteAccessServerIsUp)
		{
			restartingHaloDemoTaskShouldCancel = NO;
			[restartingHaloDemoTask cancel];
		}
		else
		{
			[NSApp runModalSession:restartingHaloServerSession];
			[restartingHaloServerProgressIndicator setDoubleValue:theProgress];
			
			if (remoteAccessServerIsUp && interestedClientSocket)
			{
				NSString *message = [NSString stringWithFormat:@"hp%i\n", theProgress];
				send(interestedClientSocket, [message UTF8String], strlen([message UTF8String]), 0);
			}
		}
	}
}

- (IBAction)syncToHaloDemo:(id)sender
{
	if ([[syncButton title] isEqualToString:@"Sync"])
	{
		[syncButton setTitle:@"Cancel"];
		[syncButton setKeyEquivalent:@"\033"];
		[syncProgressIndicator setHidden:NO];
		
		// the cheat does not know how to search UTF16 characters,
		// so we'll just use two unicode characters (16 bits each) and represent them as one 32 bit integer
		
		unichar unicodeCharacters[2];
		[@"Q)" getCharacters:unicodeCharacters
					   range:NSMakeRange(0, 2)];
		
		unicodeCharacters[0] = CFSwapInt16HostToBig(unicodeCharacters[0]);
		unicodeCharacters[1] = CFSwapInt16HostToBig(unicodeCharacters[1]);
		
		Variable *value = [[Variable alloc] init];
		[value setValue:unicodeCharacters
				   size:2 * sizeof(unichar)];
		
		SearchContext *context = [[SearchContext alloc] initWithPID:haloProcessID
													 searchOperator:TCEqual
															  value:value];
		[value release];
		
		syncingTask = [[ThreadedTask alloc] initWithFunction:[context iterationFunction]
													 context:context
													delegate:self];
		
		[context release];
		
		[syncingTask run];
	}
	else
	{
		[syncButton setEnabled:NO];
		[syncingTask cancel];
	}
}

#pragma mark Remote Access

- (void)enableRemoteAccess:(BOOL)remoteAccessState
{
	if (remoteAccessState)
	{
		if (!remoteAccessServerIsUp)
		{
			remoteAccessServerIsUp = YES;
			[NSThread detachNewThreadSelector:@selector(startRemoteAccessServer:)
									 toTarget:self
								   withObject:nil];
		}
	}
	else
	{
		remoteAccessServerIsUp = NO;
	}
}

- (void)sendServerInfoToAllClients
{
	NSString *serverNameMessage = [@"sn" stringByAppendingString:[self readInServerName]];
	NSString *passwordMessage = [@"pa" stringByAppendingString:[self readInPassword]];
	NSString *maxPlayersMessage = [@"mp" stringByAppendingFormat:@"%i", [self readInMaxPlayers]];
	NSString *spectatorStateMessage = [@"ss" stringByAppendingFormat:@"%i", [makeHostASpectatorMenuItem state]];
	
	// su, meaning halo server is up
	NSString *resultingMessage = [NSString stringWithFormat:@"%@\n%@\n%@\n%@\nsu\n", serverNameMessage, passwordMessage, maxPlayersMessage, spectatorStateMessage];
	
	int socket;
	for (socket = 0; socket <= maxSocket; socket++)
	{
		if (socket != listenerSocket && FD_ISSET(socket, &masterSet))
		{
			send(socket, [resultingMessage UTF8String], strlen([resultingMessage UTF8String]), 0);
		}
	}
}

- (void)sendServerInfoToClient:(int)clientSocket
{
	NSString *serverNameMessage = [@"sn" stringByAppendingString:[self readInServerName]];
	NSString *passwordMessage = [@"pa" stringByAppendingString:[self readInPassword]];
	NSString *maxPlayersMessage = [@"mp" stringByAppendingFormat:@"%i", [self readInMaxPlayers]];
	NSString *spectatorStateMessage = [@"ss" stringByAppendingFormat:@"%i", [makeHostASpectatorMenuItem state]];
	
	NSString *resultingMessage = [NSString stringWithFormat:@"%@\n%@\n%@\n%@\nsu\n", serverNameMessage, passwordMessage, maxPlayersMessage, spectatorStateMessage];
	send(clientSocket, [resultingMessage UTF8String], strlen([resultingMessage UTF8String]), 0);
}

- (void)sendServerOffMessageToAllClients
{
	int socket;
	for (socket = 0; socket <= maxSocket; socket++)
	{
		if (socket != listenerSocket && FD_ISSET(socket, &masterSet))
		{
			send(socket, "sd", 2, 0);
		}
	}
}

- (void)startRemoteAccessServer:(id)object
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	listenerSocket = socket(AF_INET, SOCK_STREAM, 0);
	
	if (listenerSocket == -1)
	{
		perror("listener");
		
		NSRunAlertPanel(@"Server could not be started",
						@"The server failed to listen on port %i.",
						@"OK", nil, nil, [[NSUserDefaults standardUserDefaults] integerForKey:HSCListeningPort]);
		
		[self openPreferencesAndTellServerFailedToLaunch];
		[pool release];
		[NSThread exit];
	}
	
	int yes = 1;
	
	if (setsockopt(listenerSocket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1)
	{
		perror("setsockopt");
		
		NSRunAlertPanel(@"Server could not be started",
						@"The server failed to bind on port %i because the connection is alreday in use.",
						@"OK", nil, nil, [[NSUserDefaults standardUserDefaults] integerForKey:HSCListeningPort]);
		
		close(listenerSocket);
		[self openPreferencesAndTellServerFailedToLaunch];
		[pool release];
		[NSThread exit];
	}
	
	struct sockaddr_in myAddress;
	
	myAddress.sin_family = AF_INET;
	myAddress.sin_port = htons([[NSUserDefaults standardUserDefaults] integerForKey:HSCListeningPort]);
	myAddress.sin_addr.s_addr = INADDR_ANY;
	memset(myAddress.sin_zero, '\0', sizeof(myAddress.sin_zero));
	
	if (bind(listenerSocket, (struct sockaddr *)&myAddress, sizeof(myAddress)) == -1)
	{
		perror("bind");
		
		NSRunAlertPanel(@"Server could not be started",
						@"The server failed to bind on port %i.",
						@"OK", nil, nil, [[NSUserDefaults standardUserDefaults] integerForKey:HSCListeningPort]);
		
		close(listenerSocket);
		[self openPreferencesAndTellServerFailedToLaunch];
		[pool release];
		[NSThread exit];
	}
	
	const int backlog = 100;
	if (listen(listenerSocket, backlog) == -1)
	{
		perror("listen");
		
		NSRunAlertPanel(@"Server could not be started",
						@"The server failed to listen on port %i.",
						@"OK", nil, nil, [[NSUserDefaults standardUserDefaults] integerForKey:HSCListeningPort]);
		
		close(listenerSocket);
		[self openPreferencesAndTellServerFailedToLaunch];
		[pool release];
		[NSThread exit];
	}
	
	[pool release];
	
	FD_ZERO(&masterSet);
	
	FD_SET(listenerSocket, &masterSet);
	
	maxSocket = listenerSocket;
	
	while (remoteAccessServerIsUp)
	{
		pool = [[NSAutoreleasePool alloc] init];
		
		fd_set readSet = masterSet;
		
		struct timeval timeValue;
		timeValue.tv_sec = 0;
		timeValue.tv_usec = 300000;
		
		if (select(maxSocket + 1, &readSet, NULL, NULL, &timeValue) == -1)
		{
			perror("select");
			continue;
		}
		
		int socket;
		for (socket = 0; socket <= maxSocket; socket++)
		{
			if (FD_ISSET(socket, &readSet))
			{
				if (socket == listenerSocket)
				{
					// handle new connection
					
					struct sockaddr_in theirAddress;
					socklen_t sinSize = sizeof(theirAddress);
					
					int newSocket = accept(listenerSocket, (struct sockaddr *)&theirAddress, &sinSize);
					
					if (newSocket == -1)
					{
						perror("accept");
					}
					else
					{
						if (newSocket > maxSocket)
						{
							maxSocket = newSocket;
						}
						
						FD_SET(newSocket, &masterSet);
					}
				}
				else
				{
					// client sent us something
					
					const int MAX_BUFFER_SIZE = 256;
					char buffer[MAX_BUFFER_SIZE];
					int numberOfBytes = recv(socket, buffer, MAX_BUFFER_SIZE, 0);
					
					if (numberOfBytes == -1)
					{
						perror("recv");
					}
					else if (numberOfBytes == 0)
					{
						// client quit
						
						FD_CLR(socket, &masterSet);
						close(socket);
					}
					else
					{
						// we got something from the client
						
						buffer[numberOfBytes] = '\0';
						
						NSArray *messageComponents = [[NSString stringWithUTF8String:buffer] componentsSeparatedByString:@"\n"];
						NSEnumerator *messageEnumerator = [messageComponents objectEnumerator];
						NSString *message;
						
						while (message = [messageEnumerator nextObject])
						{
							if (!message || [message length] < 2)
							{
								continue;
							}
							
							if ([message characterAtIndex:0] == 'p' && [message characterAtIndex:1] == 'w')
							{
								// password
								
								if ([[message substringFromIndex:2] isEqualToString:[[NSUserDefaults standardUserDefaults] objectForKey:HSCServerPassword]])
								{
									// good password
									send(socket, "gd \n", 4, 0);
									
									// also send the client the current server info, if there's any
									if (serverIsUp)
									{
										[self sendServerInfoToClient:socket];
									}
								}
								else
								{
									// bad password
									send(socket, "bd \n", 4, 0);
								}
							}
							else if ([message characterAtIndex:0] == 's' && [message characterAtIndex:1] == 'n')
							{
								// change server name
								[serverNameTextField setStringValue:[message substringFromIndex:2]];
								[self applyChanges:nil];
							}
							else if ([message characterAtIndex:0] == 'p' && [message characterAtIndex:1] == 'a')
							{
								// change password
								[passwordTextField setStringValue:[message substringFromIndex:2]];
								[self applyChanges:nil];
							}
							else if ([message characterAtIndex:0] == 'm' && [message characterAtIndex:1] == 'p')
							{
								// change max players
								[maxPlayersTextField setStringValue:[message substringFromIndex:2]];
								[maxPlayersStepper setIntValue:[[message substringFromIndex:2] intValue]];
								[self applyChanges:nil];
							}
							else if ([message characterAtIndex:0] == 'r' && [message characterAtIndex:1] == 's')
							{
								// client wants a restore
								[self sendServerInfoToClient:socket];
							}
							else if ([message characterAtIndex:0] == 'h' && [message characterAtIndex:1] == 'r')
							{
								// client wants halo server to restart
								interestedClientSocket = socket;
								[self restartHaloDemoServer:nil];
							}
							else if ([message characterAtIndex:0] == 'h' && [message characterAtIndex:1] == 'c')
							{
								// client wants to cancel halo server restart, but do nothing if we are restarting from the address file
								if (![[NSFileManager defaultManager] fileExistsAtPath:HALO_RESTART_ADDRESSES_PATH])
								{
									interestedClientSocket = 0;
									restartingHaloDemoTaskShouldCancel = YES;
								}
							}
							else if ([message characterAtIndex:0] == 's' && [message characterAtIndex:1] == 's')
							{
								// client wants to toggle the spectator state (and I hate threads)
								hostIsGoingToSpectateFromThreadNotice = YES;
							}
						}
					}
				}
			}
		}
		
 		[pool release];
	}
	
	// close all sockets
	
	int socket;
	for (socket = 0; socket <= maxSocket; socket++)
	{
		if (socket != listenerSocket && FD_ISSET(socket, &masterSet))
		{
			FD_CLR(socket, &masterSet);
			close(socket);
		}
	}
	
	close(listenerSocket);
	
	pool = [[NSAutoreleasePool alloc] init];
	
	[[NSNotificationCenter defaultCenter] postNotificationName:HSCServerShutOffNotification
														object:nil];
	
	[pool release];
}

#pragma mark Opening Stuff

- (IBAction)openPreferences:(id)sender
{
	if (!preferencesController)
	{
		preferencesController = [[HSCPreferencesController alloc] init];
		[NSBundle loadNibNamed:@"Preferences"
						 owner:preferencesController];
	}
	
	[preferencesController showWindow];
}

- (void)openPreferencesAndTellServerFailedToLaunch
{
	[self openPreferences:nil];
	[preferencesController turnOffServer];
}

- (IBAction)openMainWindow:(id)sender
{
	if (!readServerInfoTimer)
	{
		
		readServerInfoTimer = [NSTimer scheduledTimerWithTimeInterval:1.0
															   target:self
															 selector:@selector(readServerInfo:)
															 userInfo:nil
															  repeats:YES];
	}
	
	[mainWindow makeKeyAndOrderFront:nil];
}

@end
